(*-----------------------------------------------------------------------*)
(* Program: Stream                                                       *)
(* Original code developed by John D. McCalpin and Joe R. Zagar          *)
(* Programmers: NISHIMURA Ryohei                                         *)
(*                                                                       *)
(* This program measures memory transfer rates in MB/s for simple        *)
(* computational kernels coded in OCaml.                                 *)
(*-----------------------------------------------------------------------*)
(* Copyright 2011:      NISHIMURA Ryohei                                 *)
(* Copyright 1991-2005: John D. McCalpin                                 *)
(*-----------------------------------------------------------------------*)
(* License:                                                              *)
(*  1. You are free to use this program and/or to redistribute           *)
(*     this program.                                                     *)
(*  2. You are free to modify this program for your own use,             *)
(*     including commercial use, subject to the publication              *)
(*     restrictions in item 3.                                           *)
(*  3. You are free to publish results obtained from running this        *)
(*     program, or from works that you derive from this program,         *)
(*     with the following limitations:                                   *)
(*     3a. In order to be referred to as "STREAM benchmark results",     *)
(*         published results must be in conformance to the STREAM        *)
(*         Run Rules, (briefly reviewed below) published at              *)
(*         http://www.cs.virginia.edu/stream/ref.html                    *)
(*         and incorporated herein by reference.                         *)
(*         As the copyright holder, John McCalpin retains the            *)
(*         right to determine conformity with the Run Rules.             *)
(*     3b. Results based on modified source code or on runs not in       *)
(*         accordance with the STREAM Run Rules must be clearly          *)
(*         labelled whenever they are published.  Examples of            *)
(*         proper labelling include:                                     *)
(*         "tuned STREAM benchmark results"                              *)
(*         "based on a variant of the STREAM benchmark code"             *)
(*         Other comparable, clear and reasonable labelling is           *)
(*         acceptable.                                                   *)
(*     3c. Submission of results to the STREAM benchmark web site        *)
(*         is encouraged, but not required.                              *)
(*  4. Use of this program or creation of derived works based on this    *)
(*     program constitutes acceptance of these licensing restrictions.   *)
(*  5. Absolutely no warranty is expressed or implied.                   *)
(*-----------------------------------------------------------------------*)
let stream n offset ntimes =
  let mysecond = Unix.gettimeofday in
  let hline _ = Printf.printf
      "-------------------------------------------------------------\n" in
  let () = hline () in
  let () = Printf.printf "STREAM version $Revision: 5.9 $\n" in
  let () = hline () in
  let bytes_per_word = 8 in
  let () = Printf.printf
      "This system uses %d bytes per DOUBLE PRECISION word.\n" bytes_per_word in
  let () = hline () in
  let () = Printf.printf "Array size = %d, Offset = %d\n" n offset in
  let () = Printf.printf "Each test is run %d times, but only\n" ntimes in
  let () = Printf.printf "the *best* time for each is used.\n" in
  let () = hline () in
  let (a, b, c) =
    let rec init n a b c j =
      if j >= n then
	(a, b, c)
      else
	init n (1.0 :: a) (2.0 :: b) (0.0 :: c) (j + 1) in
    init n [] [] [] 0 in
  let quantum =
    let quantum =
      let checktick _ =
	let rec timesfound m ret i =
	  if i >= m then
	    ret
	  else
	    let t1 = mysecond () in
	    let rec timesfound_iter t1 =
	      let t2 = mysecond () in
	      if t2 -. t1 < 1.0e-6 then
		timesfound_iter t1
	      else
		t2 in
	    timesfound m (timesfound_iter t1 :: ret) (i + 1) in
	let rec delta ret timesfound =
	  match timesfound with
	    [] -> ret
	  | first :: tl -> match tl with
	      [] -> ret
	    | second :: _ ->
		delta (int_of_float (1.0e6 *. (first -. second)) :: ret) tl in
	List.fold_left Pervasives.min Pervasives.min_int
	  (delta [] (timesfound 20 [] 0)) in
      checktick () in
    if quantum >= 1 then
      let () =
	Printf.printf "Your clock granularity/precision appears to be " in
      let () = Printf.printf "%d microseconds.\n" quantum in
      quantum
    else
      let () = Printf.printf "Your clock granularity appears to be " in
      let () = Printf.printf "less than one microsecond.\n" in
      1 in
  let t = mysecond () in
  let a =
    let rec test newa olda =
      match olda with
	[] -> newa
      | hd :: tl -> test (2.0e0 *. hd :: newa) tl in
    test [] a in
  let t = 1.0e6 *. (mysecond () -. t) in
  let () = Printf.printf "Each test below will take on the order" in
  let () = Printf.printf " of %d microseconds.\n" (int_of_float t) in
  let () = Printf.printf "   (= %d clock ticks)\n"
      (int_of_float (t /. float_of_int(quantum))) in
  let () =
    Printf.printf "Increase the size of the arrays if this shows that\n" in
  let () =
    Printf.printf "you are not getting at least 20 clock ticks per test.\n" in
  let () = hline () in
  let () = Printf.printf "WARNING -- The above is only a rough guideline.\n" in
  let () = Printf.printf "For best results, please be sure you know the\n" in
  let () = Printf.printf "precision of your system timer.\n" in
  let () = hline () in
  let (times, a, b, c) =
    let rec loop scalar k a b c ret =
      if k = 0 then
	(ret, a, b, c)
      else
	let times0k = mysecond () in
	let c =
	  let rec test0 c a =
	    match a with
	      [] -> c
	    | hd :: tl -> test0 (hd :: c) tl in
	  test0 [] a in
	let times0k = mysecond () -. times0k in
	let times1k = mysecond () in
	let b =
	  let rec test1 b c =
	    match c with
	      [] -> b
	    | hd :: tl -> test1 (scalar *. hd :: b) tl in
	  test1 [] c in
	let times1k = mysecond () -. times1k in
	let times2k = mysecond () in
	let c =
	  let rec test2 c a b =
	    match a with
	      [] -> c
	    | hd :: tl -> test2 (hd +. List.hd b :: c) tl (List.tl b) in
	  test2 [] a b in
	let times2k = mysecond () -. times2k in
	let times3k = mysecond () in
	let a =
	  let rec test3 a b c =
	    match b with
	      [] -> a
	    | hd :: tl ->
		test3 (hd +. scalar *. List.hd c :: a) tl (List.tl c) in
	  test3 [] b c in
	let times3k = mysecond () -. times3k in
	loop scalar (k - 1) a b c
	  ([times0k; times1k; times2k; times3k] :: ret) in
    loop 3.0 ntimes a b c [] in
  let rec foldfold f a b =
    let rec g f a b =
      match a with
	[] -> []
      | hd :: tl -> f hd (List.hd b) :: g f tl (List.tl b) in
    match b with
      [] -> a
    | hd :: tl -> match tl with
	[] -> a
      | _ :: _ -> foldfold f (g f a hd) tl in
  let avgtime = foldfold (+.) [0.0; 0.0; 0.0; 0.0] times in
  let mintime = foldfold Pervasives.min
      [Pervasives.infinity; Pervasives.infinity;
       Pervasives.infinity; Pervasives.infinity]
      times in
  let maxtime = foldfold Pervasives.max
      [Pervasives.neg_infinity; Pervasives.neg_infinity;
       Pervasives.neg_infinity; Pervasives.neg_infinity]
      times in
  let () = Printf.printf
      "Function      Rate (MB/s)   Avg time     Min time     Max time\n" in
  let () =
    let rec print ntimes label bytes avgtime mintime maxtime =
      match label with
	[] -> ()
      | hd :: tl ->
	  let avgtimej = (List.hd avgtime) /. float_of_int (ntimes - 1) in
	  let () = Printf.printf "%s%11.4f  %11.4f  %11.4f  %11.4f\n" hd
	      (1.0e-06 *. (List.hd bytes) /. (List.hd mintime))
	      avgtimej
	      (List.hd mintime)
	      (List.hd maxtime) in
	  print ntimes tl (List.tl bytes)
	    (List.tl avgtime) (List.tl mintime) (List.tl maxtime) in
    print ntimes
      ["Copy:      "; "Scale:     "; "Add:       "; "Triad:     "]
      [float_of_int (2 * bytes_per_word * n);
       float_of_int (2 * bytes_per_word * n);
       float_of_int (3 * bytes_per_word * n);
       float_of_int (3 * bytes_per_word * n)]
      avgtime mintime maxtime in
  let () = hline () in
  let check_stream_results n ntimes a b c =
    let aj = 1.0 in
    let bj = 2.0 in
    let cj = 0.0 in
    let aj = 2.0e0 *. aj in
    let (aj, bj, cj) =
      let rec loop scalar aj bj cj k =
	if k = 0 then
	  (aj, bj, cj)
	else
	  let cj = aj in
	  let bj = scalar *. cj in
	  let cj = aj +. bj in
	  let aj = bj +. scalar *. cj in
	  loop scalar aj bj cj (k - 1) in
      loop 3.0 aj bj cj ntimes in
    let aj = aj *. float_of_int n in
    let bj = bj *. float_of_int n in
    let cj = cj *. float_of_int n in
    let asum = List.fold_left (+.) 0.0 a in
    let bsum = List.fold_left (+.) 0.0 b in
    let csum = List.fold_left (+.) 0.0 c in
    let epsilon  = 1.0e-8 in
    if Pervasives.abs_float (aj -. asum) /. asum > epsilon then
      let () = Printf.printf "Failed Validation on array a[]\n" in
      let () = Printf.printf "        Expected  : %f \n"aj in
      Printf.printf "        Observed  : %f \n" asum
    else if Pervasives.abs_float (bj -. bsum) /. bsum > epsilon then
      let () = Printf.printf "Failed Validation on array a[]\n" in
      let () = Printf.printf "        Expected  : %f \n"bj in
      Printf.printf "        Observed  : %f \n" bsum
    else if Pervasives.abs_float (cj -. csum) /. csum > epsilon then
      let () = Printf.printf "Failed Validation on array a[]\n" in
      let () = Printf.printf "        Expected  : %f \n"cj in
      Printf.printf "        Observed  : %f \n" csum
    else
      Printf.printf "Solution Validates\n" in
  let () = check_stream_results n ntimes a b c in
  hline () in
stream 2000000 0 10;;
